/*
** Module   : DisplayFormat.java
** Abstract : display formatting base class.
**
** Copyright (c) 2005-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description---------------------
** 001 SVG 20050609   @21446 Initial version.
** 002 SVG 20050620   @21523 DisplayFormat related classes are redesigned.
**                           Now cursor's position and insert mode are
**                           maintained inside Presentation. This allows
**                           to not pass them by reference into
**                           Presentation edit methods. All
**                           com.goldencode.p2j.ui.chui.dataref package
**                           is obsoleted and removed.
** 003 SIY 20050901   @22532 Added instanceOfType() method which generates
**                           new BaseDataType instance using given type 
**                           name and value strings.
** 004 SIY 20050831   @22533 Refactoring (see @@22394-22409) 
** 005 SIY 20050909   @22665 Fixed variables initialization.
** 006 SIY 20050901   @22532 Added instanceOfType() method which generates
**                           new BaseDataType instance using given type 
**                           name and default value.
** 007 SIY 20050919   @22835 Fixed handling of type names of different 
**                           case.
** 008 GES 20060118   @23977 JavaDoc updates.
** 009 GES 20060211   @25545 Replaced a stupid, unsafe implementation
**                           of generateMultichar() with a safe version.
** 010 GES 20060503   @25955 Use unknown value instead of 01/01/2005 as
**                           the default date.  Added a flag for edit 
**                           mode to allow differential string generation 
**                           in subclasses as needed.
** 011 GES 20060505   @25991 Added a way to query the last valid cursor
**                           position.
** 012 GES 20060517   @26202 Exposed additional interfaces for proper
**                           numeric data formatting.
** 013 GES 20060711   @27974 New method for proper fixed numeric data
**                           formatting.
** 014 GES 20060920   @26727 Added support for isUnknown().
** 015 SVL 20090225   @41372 Added support for recid type.
** 016 SIY 20090701   @43017 Added Presentation.validateFormat() method.
** 017 SIY 20090721   @43287 Common activate() implementation is moved here.
**                           Added getInitialCursorPos() method.
** 018 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 019 OM  20121218          Added support for int64 data type.
*/

package com.goldencode.p2j.ui.client.format;

import java.util.*;

import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager.ErrorEntry;

/**
 * A descendant of this class encapsulates info about Progress display format.
 * A concrete descendant corresponding to data type should be used. Class
 * <code>DisplayFormatFactory</code> is provided to make this selection
 * automatic.
 * <p>
 * After format instantiation <code>fromVar()</code> method may be called to
 * get {@link DisplayFormat.Presentation an internal presentation}of data of
 * corresponding data type.
 */
public abstract class DisplayFormat
{
   /** Saved Progress format definition string. */
   private String formatDef;

   /**
    * Constructs format from the Progress format string. Only saves this
    * string. Format dependent parser is placed to constructor of this class
    * descendant.
    * 
    * @param   format
    *          Progress format definition string.
    */
   public DisplayFormat(String format)
   {
      formatDef = format;
   }

   /**
    * Generates string from question marks.
    * 
    * @param   len
    *          Requested result string length.
    * 
    * @return  String from question marks with given length.
    */
   public static String generateQuestions(int len)
   {
      return generateMultichar('?', len);
   }

   /**
    * Generates string from spaces.
    * 
    * @param   len
    *          Requested result string length.
    * 
    * @return  String from spaces with given length.
    */
   public static String generateSpaces(int len)
   {
      return generateMultichar(' ', len);
   }

   /**
    * Formats screen value into screen chars. This includes scrolling, padding
    * and truncating if necessary.
    * 
    * @param   screen
    *          Screen value or column's title. May contain <code>null</code>.
    *          In this case return value will consists from question marks.
    * @param   width
    *          Width on display. The result string will have this length.
    * @param   offsScroll
    *          Scroll offset. Display width plus this offset must be no more
    *          than screen value length. Used only when screen value is longer
    *          than display width and left trimming is not requested.
    * @param   leftTrim
    *          Indicates that screen value should be truncated from the left.
    *          Used only when screen value is longer than display width.
    * @param   rightAlign
    *          Indicates that short screen value should be padded on the left.
    * 
    * @return  String with <code>width</code> length which can be placed to
    *          terminal.
    */
   public static String formatScreenValue(String  screen,
                                          int     width,
                                          int     offsScroll,
                                          boolean leftTrim,
                                          boolean rightAlign)
   {
      if (screen == null)
      {
         return generateQuestions(width);
      }
      
      int lenDiff = screen.length() - width;
      
      if (lenDiff == 0)
      {
         return screen;
      }
      else if (lenDiff < 0)
      {
         StringBuffer sb = new StringBuffer(width);
         
         if (!rightAlign)
         {
            sb.append(screen);
         }
         
         for (int i = lenDiff; i < 0; i++)
         {
            sb.append(' ');
         }
         
         if (rightAlign)
         {
            sb.append(screen);
         }
         
         return sb.toString();
      }
      
      // screen value is longer than width
      if (leftTrim)
      {
         // left trim it
         return screen.substring(lenDiff);
      }
      
      // right trimming with scrolling
      return screen.substring(offsScroll, offsScroll + width);
   }

   /**
    * Returns format definition string.
    * 
    * @return  Format definition string.
    */
   public String getFormatDef()
   {
      return formatDef;
   }

   /**
    * Returns screen width of the format.  The screen width is the number of
    * characters that will be rendered.  This is used as the default size for
    * an editor.
    * 
    * @return   The screen width of the format. 
    */
   public abstract int getScreenWidth();

   /**
    * Indicates if the screen value or a column's title could be formatted
    * with trimming from the left and right alignment based on this format.
    * <p>
    * This value only notes if this format supports such trimming/alignment
    * but it doesn't specify if it should be applied in every case (which
    * may be a configuration option for a given caller).
    * 
    * @return   Always <code>false</code>. 
    */
   public boolean isRightPositioning()
   {
      return false;
   }

   /**
    * Indicates that ZAP operation can be performed on field with this format.
    * 
    * @return  Is ZAP should be performed for this format when it is possible
    *          (AUTO-ZAP set etc.).
    */
   public boolean isZapSupported()
   {
      return true;
   }

   /**
    * Returns a new presentation instance with a default initialized insert
    * mode and cursor position.
    * 
    * @param   pres
    *          A Presentation instance which was constructed by this format.
    *          May be <code>null</code>. This presentation instance is not 
    *          changed in this method.
    * 
    * @return  New presentation instance.
    */
   public abstract Presentation clear(Presentation pres);

   /**
    * Initializes a presentation for variable's value.
    * 
    * @param   var
    *          Variable which should be compatible with format.
    * 
    * @return  Data presentation
    * 
    * @throws  DisplayFormatCheckException
    *          Variable's value is incompatible with format.
    */
   public abstract Presentation fromVar(BaseDataType var)
   throws DisplayFormatCheckException;

   /**
    * Fast generation of short string from repetitions of the same char.
    * 
    * @param   ch
    *          Character to duplicate.
    * @param   len
    *          Requested result string length.
    * 
    * @return  String from template chars with given length
    */
   private static String generateMultichar(char ch, int len)
   {
      char[] multi = new char[len];

      Arrays.fill(multi, ch);

      return new String(multi);       
   }

   /**
    * Generate instance of <code>BaseDataType</code> and default value.
    * <p>
    * Default value for string is empty string, 0 for numbers, false for
    * logical and <code>unknown value</code> for dates.
    * 
    * @param   dataType
    *          Data type name.
    * @return  Instance of <code>BaseDataType</code> or <code>null</code>
    *          if type name is not recognized.
    */
   public static BaseDataType instanceOfType(String dataType)
   {
      if (dataType == null)
         return null;
      
      dataType = dataType.toLowerCase();

      BaseDataType variable = null;

      if (dataType.equals("integer"))
         variable = new integer(0);
      else if (dataType.equals("int64"))
         variable = new int64(0);
      else if (dataType.equals("recid"))
         variable = new recid(0);
      else if (dataType.equals("decimal"))
         variable = new decimal(0.0);
      else if (dataType.equals("logical"))
         variable = new logical(false);
      else if (dataType.equals("date"))
         variable = date.instantiateUnknownDate();
      else if (dataType.equals("character"))
         variable = new character("");

      return variable;
   }
   
   /**
    * Generate instance of <code>BaseDataType</code> for given type name and
    * value.
    * 
    * @param   dataType
    *          Data type name.
    * @param   value
    *          Initial value for the variable.
    * @return  Instance of <code>BaseDataType</code> or <code>null</code>
    *          if type name is not recognized.
    */
   public static BaseDataType instanceOfType(String dataType, String value)
   {
      if (dataType == null)
         return null;
      
      dataType = dataType.toLowerCase();
      
      BaseDataType variable = null;

      if (dataType.equals("integer"))
         variable = new integer(value);
      else if (dataType.equals("decimal"))
         variable = new decimal(value);
      else if (dataType.equals("logical"))
      {
         if (value != null && value.length() > 0)
            variable = new logical(Boolean.valueOf(value).booleanValue());
         else
            variable = new logical();
      }
      else if (dataType.equals("date"))
         variable = new date(value);
      else if (dataType.equals("character"))
         variable = new character(value);

      return variable;
   }

   /**
    * Format specific internal presentation of data. The common usage of this
    * class is to get formatted string with <code>toScreenValue()</code>
    * call. Other methods are mainly for FillIn.
    */
   public abstract class Presentation
   {
      /**
       * Cursor's position. Usually it is changed only by input (
       * <code>insert()</code>,<code>deleteBack()</code> etc.) methods.
       * Default value is <code>0</code>.
       */
      protected int presCursorPos = 0;

      /**
       * Insert mode (<code>true</code>- insert, <code>false</code>-
       * overwrite). May be changed as from input methods as directly from a
       * widget. Default value is <code>false</code>.
       */ 
      protected boolean presInsertMode = false;
      
      /**
       * Is the last format's character entered. This flag is set by
       * <code>input()</code> method and is necessary to implement
       * AUTO-RETURN FillIn behaviour.
       */
      private boolean presLastCharEntered = false;
      
      /**
       * Simple constructor. Just a placeholder.
       */
      public Presentation()
      {
      }

      /**
       * Returns the last valid cursor position in the current data.  In most
       * cases this is the same as the screen width of this format, however
       * in some subclasses this will be equivalent to the number of actual
       * character positions in the data.
       * 
       * @return  The last valid cursor position.
       */
      public int getLastPos()
      {
         return (getScreenWidth() - 1);
      }

      /**
       * Returns the cursor's position.
       * 
       * @return  The cursor's position.
       */
      public int getCursorPos()
      {
         return presCursorPos;
      }

      /**
       * Sets the new cursor's position.
       * 
       * @param   cursorPos
       *          The new cursor's position.
       */
      public void setCursorPos(int cursorPos)
      {
         presCursorPos = cursorPos;
      }

      /**
       * Returns the sign state of the numeric value.
       *
       * @return  Always <code>false</code>.
       */
      public boolean isNegative()
      {
         return false;
      }
   
      /**
       * Detects if the numeric sign formatting character exists in the
       * format and if it is optional.
       *
       * @return   Always <code>false</code>.
       */
      public boolean isOptionalLeftSign()
      {
         return false;
      }
      
      /**
       * Detects if the numeric format is comprised of digits that always
       * output (like "-99,999").
       *
       * @return   Always <code>false</code>.
       */
      public boolean isFixedNumeric()
      {
         return false;
      }
      
      /**
       * Returns the insert mode state.
       * 
       * @return  Insert mode.
       *          <ul>
       *             <li><code>false</code>- overwrite mode,
       *             <li><code>true</code>- insert mode.
       *          </ul>
       */
      public boolean getInsertMode()
      {
         return presInsertMode;
      }

      /**
       * Changes the insert mode state.
       * 
       * @param   insertMode
       *          New state.
       *          <ul>
       *             <li><code>false</code>- overwrite mode,
       *             <li><code>true</code>- insert mode.
       *          </ul>
       */
      public void setInsertMode(boolean insertMode)
      {
         presInsertMode = insertMode;
      }

      /**
       * Returns is the last format's character entered.
       * 
       * @return  <code>true</code> if the last format's character entered.
       */
      public boolean isLastCharEntered()
      {
         return presLastCharEntered;
      }

      /**
       * Sets is the last format's character entered.
       * 
       * @param   lastCharEntered
       *          Is the last format's character entered.
       */
      public void setLastCharEntered(boolean lastCharEntered)
      {
         presLastCharEntered = lastCharEntered;
      }
      
      /**
       * Formats the data as a string that is ready for display.
       * 
       * @param    edit
       *           <code>true</code> if the screen value is being generated
       *           for interactive editing purposes.
       * 
       * @return   The formatted string.
       */
      public String toScreenValue(boolean edit)
      {
         return toScreenValue();
      }
      
      /**
       * Performs the necessary initialization for a widget which is
       * activated by a focus change (using a tab).
       */
      public void activate()
      {
         presCursorPos = getInitialCursorPos();
      }

      /**
       * Get initial cursor position which will be set during activation.
       * 
       * @return  calculated cursor position.
       */
      public int getInitialCursorPos()
      {
         return 0;
      }
      
      /**
       * Checks format rules. Method finishEdit() should be called before it.
       * 
       * @throws  DisplayFormatCheckException
       *          When data is incompatible with format
       */
      public abstract void checkFormat()
      throws DisplayFormatCheckException;

      /**
       * Validate representation according to format rules and return error
       * message if there is an error.
       * 
       * @return   <code>null</code> if no error or error message if there is an
       *           error.
       */
      public ErrorEntry validateFormat()
      {
         return null;
      }
      
      /**
       * Returns if the presentation currently represents the 
       * <code>unknown value</code>.  If any editing has occurred, then this
       * will return <code>false</code> even if the editing is not complete.
       *
       * @return   <code>true</code> if this is <code>unknown</code>.
       */
      public abstract boolean isUnknown();
      
      /**
       * Inputs character. The result depends from the current insert mode.
       * 
       * @param   ch
       *          Input character
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      public abstract boolean input(char ch);

      /**
       * Erases one symbol before the cursor position.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      public abstract boolean deleteBack();

      /**
       * Erases one symbol in the cursor position.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      public abstract boolean deleteCurrent();

      /**
       * Moves cursor one character to the left.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      public abstract boolean moveCursorLeft();

      /**
       * Moves cursor one character to the right. Also updates insert mode.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      public abstract boolean moveCursorRight();

      /**
       * Moves cursor to the beginning of the field.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      public abstract boolean moveCursorBegin();

      /**
       * Moves cursor to the end of the field.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      public abstract boolean moveCursorEnd();

      /**
       * Indicates completion of an edit. The presentation instance may fill 
       * some unfilled places.
       * 
       * @return   <code>true</code> if rendering is necessary.
       */
      public abstract boolean finishEdit();

      /**
       * Formats the data as a string that is ready for display.
       * 
       * @return   The formatted string.
       */
      public abstract String toScreenValue();

      /**
       * Instantiates a variable that is equivalent to the presentation's
       * internal representation.
       * 
       * @param    var
       *           A variable which must be compatible with the format.
       */
      public abstract void toVar(BaseDataType var);
   }
}
