/*
** Module   : MathOps.java
** Abstract : Progress 4GL compatible MathOps object
**
** Copyright (c) 2004-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050519   @21249 Created initial version with full support
**                           for all Progress math processing including
**                           all functions and arithmetic operators. 
** 002 GES 20050608   @21463 Minor fixes and additions.
** 003 GES 20050830   @22386 Added error condition processing.
** 004 GES 20060123   @24018 Minor JavaDoc change.
** 005 GES 20060412   @25490 Added negate() signatures.
** 006 NVS 20060605   @26955 Fixed divide() method when it returns an
**                           unknown value.
** 007 GES 20070321   @32507 Javadoc improvements.
** 008 ECF 20070608   @34010 Implemented decimal math using BigDecimal
**                           instead of double wherever possible. Methods
**                           which use double are maintained for
**                           compatibility.
** 009 ECF 20070618   @34091 Added a variant of round() which accepts a
**                           BigDecimal value and a precision.
** 010 ECF 20070627   @34281 Removed improper divide() optimization for
**                           integers. In cases where both operands are
**                           integers, we were doing integer division,
**                           when we needed to do fractional division.
**                           Also modified modulo() implementation to use
**                           BigDecimal.
** 011 WCP 20071008   @35397 Modulo implementation in 4GL has a peculiar
**                           behavior (incorrect) when a negative number
**                           is given as the first operand. Both the
**                           primitive and BigDecimal forms of the
**                           operator needed to be updated to be
**                           compatible.
** 012 ECF 20071008   @35398 Re-implemented pow(NumberType, NumberType) to
**                           use BigDecimal when possible. Synchronized
**                           access to random number generator. Optimized
**                           modulo(NumberType, NumberType) slightly by
**                           re-using decimal.ZERO constant.
** 013 ECF 20080306   @37463 Fixed defect in random(). Computed result was
**                           able to go beyond low end of range by one.
** 014 ECF 20080330   @37729 Added optimized variants of modulo(). In the
**                           common case of integral parameters, we avoid
**                           the considerable overhead of BigDecimal.
** 015 GES 20080415   @38017 Changed to BigDecimal(int) constructor
**                           instead of using BigDecimal(String) for the
**                           same case. This constructor is new and should
**                           be more performant.
** 016 GES 20080603   @38522 Fix for a certain plus() case which was
**                           doubling operand 2 instead of adding the
**                           two operands.
** 017 GES 20090422   @41911 Converted to standard string formatting.
** 018 SIY 20100408   @44803 Fixed minus() method.
** 019 OM  20121219          Added support for int64 ops.
*/

package com.goldencode.p2j.util;

import java.math.*;
import java.util.Random;

/**
 * A class that provides Progress 4GL compatible math functions and arithmetic
 * operators.  This class provides all functionality as statics and has no
 * instance data.
 * <p>
 * The following is the mapping of Progress language features to the
 * corresponding feature in this class:
 * <p>
 * <pre>
 * +        operator                 {@link #plus(NumberType,NumberType)}
 * +        operator                 {@link #plus(integer,integer)}
 * -        operator                 {@link #minus(NumberType,NumberType)}
 * -        operator                 {@link #minus(integer,integer)}
 * -        operator (unary minus)   {@link #negate(decimal)}
 * -        operator (unary minus)   {@link #negate(integer)}
 * *        operator                 {@link #multiply(NumberType,NumberType)}
 * *        operator                 {@link #multiply(integer,integer)}
 * /        operator                 {@link #divide(NumberType,NumberType)}
 * MODULO   operator                 {@link #modulo(NumberType,NumberType)}
 * absolute function                 {@link #abs(decimal)}
 * absolute function                 {@link #abs(integer)}
 * log      function                 {@link #log(NumberType,NumberType)}
 * exp      function                 {@link #pow(NumberType,NumberType)}
 * random   function                 {@link #random(integer,integer)}
 * round    function                 {@link #round(decimal,integer)} 
 * sqrt     function                 {@link #sqrt(NumberType)}
 * truncate function                 {@link #truncate(decimal,integer)}
 * </pre>
 * <p>
 * The above methods all return an instance of {@link decimal} or 
 * {@link integer} which allows the Progress unknown value to be properly
 * handled.  Overloaded versions of the operator and function methods exist
 * which handle different combinations of primative type parameters, allowing
 * the resulting code to be implemented without constant construction of
 * the wrapper types for all parameters.
 * <p>
 * This class implements unknown value logic for all operators and functions.
 * Generally, if any parameter of the these methods is unknown, the result is
 * unknown.
 * <p>
 * TODO:  need BigDecimal-based implementations of log() and sqrt() methods.
 * Current implementations still use floating point algorithms, which have
 * potential precision problems.
 *
 * @author    GES
 */
public final class MathOps
{
   /** Largest mantissa that can be safely represented without digit loss. */
   private static final double MAX_SAFE = 17976931348623157.0D;
   
   /**
    * Cached random number generator.
    * <p>
    * Note: this is safe (and actually it is preferred) to be JVM-wide 
    * (static) rather than a context-local instance.  This is true since 
    * any use of this from multiple threads will have a much more random
    * result than if the usage was only on one thread.
    */
   private static Random generator = null;
   
   /**
    * Return the absolute value of the passed numeric parameter.
    *
    * @param    d
    *           The number on which to take the absolute value.
    *
    * @return   The resulting absolute value or <code>unknown value</code>
    *           when the input is the <code>unknown value</code>.
    */
   public static decimal abs(decimal d)
   {
      if (d.isUnknown())
         return (decimal) d.instantiateUnknown();
      
      return (new decimal(d.toBigDecimal().abs()));
   }
   /**
    * Return the absolute value of the passed numeric parameter.
    *
    * @param    d
    *           The number on which to take the absolute value.
    *
    * @return   The resulting absolute value.
    */
   public static decimal abs(double d)
   {
      return new decimal(Math.abs(d));      
   }
   /**
    * Return the absolute value of the passed numeric parameter.
    *
    * @param    i
    *           The number on which to take the absolute value.
    *
    * @return   The resulting absolute value or <code>unknown value</code>
    *           when the input is the <code>unknown value</code>.
    */
   public static integer abs(integer i)
   {
      if (i.isUnknown())
         return (integer) i.instantiateUnknown();
      
      return abs(i.intValue());
   }
   /**
    * Return the absolute value of the passed numeric parameter.
    *
    * @param    i
    *           The number on which to take the absolute value.
    *
    * @return   The resulting absolute value.
    */
   public static integer abs(int i)
   {
      return new integer(Math.abs(i));
   }
   /**
    * Return the absolute value of the passed int64 parameter.
    *
    * @param    i
    *           The number on which to take the absolute value.
    *
    * @return   The resulting absolute value or <code>unknown value</code>
    *           when the input is the <code>unknown value</code>.
    */
   public static int64 abs(int64 i)
   {
      if (i.isUnknown())
         return (int64) i.instantiateUnknown();
      
      return abs(i.longValue());
   }
   /**
    * Return the absolute value of the passed numeric parameter.
    *
    * @param    i
    *           The number on which to take the absolute value.
    *
    * @return   The resulting absolute value.
    */
   public static int64 abs(long i)
   {
      return new int64(Math.abs(i));
   }

   /**
    * Calculate the natural logarithm of a number (base e).
    *
    * @param    num
    *           The number for which the logarithm should be computed.
    *
    * @return   The logarithmic value or the <code>unknown value</code>
    *           when either input is the <code>unknown value</code>. 
    */
   public static decimal log(NumberType num)
   {
      if (num.isUnknown())
         return new decimal();
      
      return log(num.doubleValue());      
   }
   /**
    * Calculate the logarithm of a number in a specific base.
    *
    * @param    num
    *           The number for which the logarithm should be computed.
    * @param    base
    *           The logarithm base.
    *
    * @return   The logarithmic value or the <code>unknown value</code>
    *           when any input is the <code>unknown value</code> or when 
    *           the base is a negative number.
    */
   public static decimal log(NumberType num, NumberType base)
   {
      if (num.isUnknown() || base.isUnknown() || base.doubleValue() < 0)
         return new decimal();
      
      return log(num.doubleValue(), base.doubleValue());      
   }
   /**
    * Calculate the logarithm of a number in a specific base.
    *
    * @param    num
    *           The number for which the logarithm should be computed.
    * @param    base
    *           The logarithm base.
    *
    * @return   The logarithmic value or the <code>unknown value</code>
    *           when any input is the <code>unknown value</code> or when 
    *           the base is a negative number.
    */
   public static decimal log(NumberType num, double base)
   {
      if (num.isUnknown() || base < 0)
         return new decimal();
      
      return log(num.doubleValue(), base);      
   }
   /**
    * Calculate the logarithm of a number in a specific base.
    *
    * @param    num
    *           The number for which the logarithm should be computed.
    * @param    base
    *           The logarithm base.
    *
    * @return   The logarithmic value or the <code>unknown value</code>
    *           when any input is the <code>unknown value</code> or when 
    *           the base is a negative number.
    */
   public static decimal log(NumberType num, int base)
   {
      if (num.isUnknown() || base < 0)
         return new decimal();
      
      return log(num.doubleValue(), base);      
   }
   /**
    * Calculate the logarithm of a number in a specific base.
    *
    * @param    num
    *           The number for which the logarithm should be computed.
    * @param    base
    *           The logarithm base.
    *
    * @return   The logarithmic value or the <code>unknown value</code> when 
    *           the base is a negative number.
    */
   public static decimal log(double num, double base)
   {
      if (base < 0)
         return new decimal();
      
      return new decimal(Math.log(num)/Math.log(base));      
   }
   /**
    * Calculate the natural logarithm of a number (base e).
    *
    * @param    num
    *           The number for which the logarithm should be computed.
    *
    * @return   The logarithmic value. 
    */
   public static decimal log(double num)
   {
      return new decimal(Math.log(num));      
   }
   
   /**
    * Raise one number (the base) to the power of the a second number (the
    * exponent).
    *
    * @param    base
    *           The number to be exponentiated.
    * @param    exponent
    *           The power to which the number should be raised.
    *
    * @return   The exponentiated value or <code>unknown value</code>
    *           when either input is the <code>unknown value</code>. 
    */
   public static decimal pow(NumberType base, NumberType exponent)
   {
      if (base.isUnknown() || exponent.isUnknown())
         return new decimal();
      
      if (!(base instanceof decimal) || !(exponent instanceof integer))
      {
         // Fall back to floating point algorithm if BigDecimal implementation
         // is unnecessary (base is not a decimal) or impossible (exponent is
         // not integral).
         return pow(base.doubleValue(), exponent.doubleValue());
      }
      
      int e = ((integer) exponent).intValue();
      BigDecimal b = ((decimal) base).toBigDecimal();
      MathContext mc = new MathContext(decimal.MAX_SCALE);
      BigDecimal result = b.pow(e, mc);
      
      return (new decimal(result));
   }
   /**
    * Raise one number (the base) to the power of the a second number (the
    * exponent).
    *
    * @param    base
    *           The number to be exponentiated.
    * @param    exponent
    *           The power to which the number should be raised.
    *
    * @return   The exponentiated value or <code>unknown value</code>
    *           when either input is the <code>unknown value</code>. 
    */
   public static decimal pow(double base, NumberType exponent)
   {
      if (exponent.isUnknown())
         return new decimal();
      
      return pow(base, exponent.doubleValue());      
   }
   /**
    * Raise one number (the base) to the power of the a second number (the
    * exponent).
    *
    * @param    base
    *           The number to be exponentiated.
    * @param    exponent
    *           The power to which the number should be raised.
    *
    * @return   The exponentiated value or <code>unknown value</code>
    *           when either input is the <code>unknown value</code>. 
    */
   public static decimal pow(NumberType base, double exponent)
   {
      if (base.isUnknown())
         return new decimal();
      
      return pow(base.doubleValue(), exponent);      
   }
   /**
    * Raise one number (the base) to the power of the a second number (the
    * exponent).
    *
    * @param    base
    *           The number to be exponentiated.
    * @param    exponent
    *           The power to which the number should be raised.
    *
    * @return   The exponentiated value.
    */
   public static decimal pow(double base, double exponent)
   {
      return new decimal(Math.pow(base, exponent));      
   }

   /**
    * Generate a psuedo-randomly distributed number between the low and
    * high values inclusively.
    *
    * @param    low
    *           The bottom end of the range (inclusive).
    * @param    high
    *           The top end of the range (inclusive).
    *
    * @return   The random integral value within the specified range.
    */
   public static integer random(int low, int high)
   {
      // initialize the generator if needed
      if (generator == null)
      {
         synchronized (MathOps.class)
         {
            if (generator == null)
            {
               generator = new Random();
            }
         }
      }
      
      int offset = (high - low) + 2;
      int result;
      synchronized (generator)
      {
         result = generator.nextInt(offset) + low;
      }
      
      return new integer(result);
   }
   /**
    * Generate a psuedo-randomly distributed number between the low and
    * high values inclusively.
    *
    * @param    low
    *           The bottom end of the range (inclusive).
    * @param    high
    *           The top end of the range (inclusive).
    *
    * @return   The random integral value within the specified range.
    */
   public static integer random(int low, integer high)
   {
      if (high.isUnknown())
         return new integer();
            
      return random(low, high.intValue());
   }



   /**
    * Generate a psuedo-randomly distributed number between the low and
    * high values inclusively.
    *
    * @param    low
    *           The bottom end of the range (inclusive).
    * @param    high
    *           The top end of the range (inclusive).
    *
    * @return   The random integral value within the specified range.
    */
   public static integer random(integer low, int high)
   {
      if (low.isUnknown())
         return new integer();
            
      return random(low.intValue(), high);
   }

   /**
    * Generate a psuedo-randomly distributed number between the low and
    * high values inclusively.
    *
    * @param    low
    *           The bottom end of the range (inclusive).
    * @param    high
    *           The top end of the range (inclusive).
    *
    * @return   The random integral value within the specified range.
    */
   public static integer random(integer low, integer high)
   {
      if (low.isUnknown() || high.isUnknown())
         return new integer();
            
      return random(low.intValue(), high.intValue());
   }
   /**
    * Generate a psuedo-randomly distributed number between the low and
    * high values inclusively.
    *
    * @param    low
    *           The bottom end of the range (inclusive).
    * @param    high
    *           The top end of the range (inclusive).
    *
    * @return   The random integral value within the specified range.
    */
   public static int64 random(int64 low, integer high)
   {
      if (low.isUnknown() || high.isUnknown())
         return new int64();
            
      return random(low.longValue(), high.longValue());
   }
   /**
    * Generate a psuedo-randomly distributed number between the low and
    * high values inclusively.
    *
    * @param    low
    *           The bottom end of the range (inclusive).
    * @param    high
    *           The top end of the range (inclusive).
    *
    * @return   The random integral value within the specified range.
    */
   public static int64 random(integer low, int64 high)
   {
      if (low.isUnknown() || high.isUnknown())
         return new int64();
            
      return random(low.longValue(), high.longValue());
   }
   /**
    * Generate a psuedo-randomly distributed number between the low and
    * high values inclusively.
    *
    * @param    low
    *           The bottom end of the range (inclusive).
    * @param    high
    *           The top end of the range (inclusive).
    *
    * @return   The random integral value within the specified range.
    */
   public static int64 random(integer low, long high)
   {
      if (low.isUnknown())
         return new int64();
            
      return random(low.longValue(), high);
   }
   /**
    * Generate a psuedo-randomly distributed number between the low and
    * high values inclusively.
    *
    * @param    low
    *           The bottom end of the range (inclusive).
    * @param    high
    *           The top end of the range (inclusive).
    *
    * @return   The random integral value within the specified range.
    */
   public static int64 random(long low, integer high)
   {
      if (high.isUnknown())
         return new int64();
            
      return random(low, high.longValue());
   }
   /**
    * Generate a psuedo-randomly distributed number between the low and
    * high values inclusively.
    *
    * @param    low
    *           The bottom end of the range (inclusive).
    * @param    high
    *           The top end of the range (inclusive).
    *
    * @return   The random integral value within the specified range.
    */
   public static int64 random(long low, int64 high)
   {
      if (high.isUnknown())
         return new int64();
            
      return random(low, high.longValue());
   }
   /**
    * Generate a psuedo-randomly distributed number between the low and
    * high values inclusively.
    *
    * @param    low
    *           The bottom end of the range (inclusive).
    * @param    high
    *           The top end of the range (inclusive).
    *
    * @return   The random integral value within the specified range.
    */
   public static int64 random(int64 low, long high)
   {
      if (low.isUnknown())
         return new int64();
            
      return random(low.longValue(), high);
   }
   /**
    * Generate a psuedo-randomly distributed number between the low and
    * high values inclusively.
    *
    * @param    low
    *           The bottom end of the range (inclusive).
    * @param    high
    *           The top end of the range (inclusive).
    *
    * @return   The random integral value within the specified range.
    */
   public static int64 random(int64 low, int64 high)
   {
      if (low.isUnknown() || high.isUnknown())
         return new int64();
            
      return random(low.longValue(), high.longValue());
   }
   /**
    * Generate a psuedo-randomly distributed number between the low and
    * high values inclusively.
    *
    * @param    low
    *           The bottom end of the range (inclusive).
    * @param    high
    *           The top end of the range (inclusive).
    *
    * @return   The random integral value within the specified range.
    */
   public static int64 random(long low, long high)
   {
      // initialize the generator if needed
      if (generator == null)
      {
         synchronized (MathOps.class)
         {
            if (generator == null)
            {
               generator = new Random();
            }
         }
      }
      
      long interval = (high - low) + 1;
      long result;
      synchronized (generator)
      {
         result = generator.nextLong() % interval + low;
      }
      
      return new int64(result);
   }

   /**
    * Calculate the square root of a number.
    *
    * @param    num
    *           The number for which to obtain the square root.
    *
    * @return   The square root or <code>unknown value</code>
    *           when either input is the <code>unknown value</code>. 
    */
   public static decimal sqrt(NumberType num)
   {
      if (num.isUnknown() || num.doubleValue() < 0)
         return new decimal();
      
      return sqrt(num.doubleValue());      
   }
   /**
    * Calculate the square root of a number.
    *
    * @param    d
    *           The number for which to obtain the square root.
    *
    * @return   The square root. 
    */
   public static decimal sqrt(double d)
   {
      return new decimal(Math.sqrt(d));      
   }

   /**
    * Rounds the passed <code>decimal</code> to the user-specified number of
    * digits of precision to the right side of the decimal point.  If the 
    * precision is specified as X then the X+1th digit is 5 or more, the
    * Xth digit is increased by 1, otherwise it is left alone. All digits
    * after the Xth are then dropped/truncated.  The specified precision
    * must be non-negative and any value greater than 10 is forced to 10.
    * <p>
    * If this instance represents the <code>unknown value</code>, the
    * result will be the <code>unknown value</code>.
    *
    * @param    value
    *           The number to round.
    * @param    precision
    *           The number of digits of precision to support.
    *
    * @return   The rounded <code>decimal</code> result.
    */
   public static decimal round(decimal value, integer precision)
   {
      return new decimal(round(value.toBigDecimal(), precision.intValue()));
   }
   /**
    * Rounds the passed <code>decimal</code> to the user-specified number of
    * digits of precision to the right side of the decimal point.  If the 
    * precision is specified as X then the X+1th digit is 5 or more, the
    * Xth digit is increased by 1, otherwise it is left alone. All digits
    * after the Xth are then dropped/truncated.  The specified precision
    * must be non-negative and any value greater than 10 is forced to 10.
    * <p>
    * If this instance represents the <code>unknown value</code>, the
    * result will be the <code>unknown value</code>.
    *
    * @param    value
    *           The number to round.
    * @param    precision
    *           The number of digits of precision to support.
    *
    * @return   The rounded <code>decimal</code> result.
    */
   public static decimal round(decimal value, int precision)
   {
      return new decimal(round(value.toBigDecimal(), precision));
   }
   /**
    * Rounds the passed <code>decimal</code> to 10 digits of precision to the
    * right side of the decimal point.  If the 11th digit is 5 or more, the
    * 10th digit is increased by 1, otherwise it is left alone. All digits
    * after the 10th are then dropped/truncated.
    * <p>
    * If this instance represents the <code>unknown value</code>, the
    * result will be the <code>unknown value</code>.
    *
    * @param    value
    *           The number to round.
    *
    * @return   The rounded <code>decimal</code> result.
    */
   public static decimal round(decimal value)
   {
      return new decimal(round(value.toBigDecimal(), decimal.MAX_SCALE));
   }
   /**
    * Rounds the passed <code>double</code> to the user-specified number of
    * digits of precision to the right side of the decimal point.  If the 
    * precision is specified as X then the X+1th digit is 5 or more, the
    * Xth digit is increased by 1, otherwise it is left alone. All digits
    * after the Xth are then dropped/truncated.  The specified precision
    * must be non-negative and any value greater than 10 is forced to 10.
    * <p>
    * If the value is <code>NaN</code> the resulting <code>double</code> will
    * be set to <code>NaN</code>.
    *
    * @param    value
    *           The number to round.
    * @param    precision
    *           The number of digits of precision to support.
    *
    * @return   The rounded <code>double</code> result.
    */
   public static double round(double value, int precision)
   {
      return truncateWorker(value, precision, true);
   }
   /**
    * Rounds the given decimal value to the user-specified number of digits
    * of precision to the right side of the decimal point.  If the precision 
    * is specified as X then the X+1th digit is 5 or more, the Xth digit is
    * increased by 1, otherwise it is left alone. All digits after the Xth
    * are then dropped/truncated.  The specified precision must be
    * non-negative and any value greater than 10 is forced to 10.
    * <p>
    * If the value is <code>null</code>, <code>null</code> is returned.  If
    * precision is negative, 0 is returned.
    *
    * @param    value
    *           The number to round.
    * @param    precision
    *           The number of digits of precision to support.
    *
    * @return   The rounded <code>decimal</code> result.
    */
   public static BigDecimal round(BigDecimal value, int precision)
   {
      return round(value, precision, BigDecimal.ROUND_HALF_UP);
   }
   /**
    * Rounds the given decimal value to 10 digits of precision to the right
    * side of the decimal point.  If the 11th digit is 5 or more, the 10th
    * digit is increased by 1, otherwise it is left alone. All digits after
    * the 10th are then dropped/truncated.
    * <p>
    * If the value is <code>null</code>, <code>null</code> is returned.
    *
    * @param    value
    *           The number to round.
    *
    * @return   The rounded <code>decimal</code> result.
    */
   public static BigDecimal round(BigDecimal value)
   {
      return round(value, decimal.MAX_SCALE, BigDecimal.ROUND_HALF_UP);
   }
   /**
    * Rounds the passed <code>double</code> to 10 digits of precision to the
    * right side of the decimal point.  If the 11th digit is 5 or more, the
    * 10th digit is increased by 1, otherwise it is left alone. All digits
    * after the 10th are then dropped/truncated.
    * <p>
    * If the value is <code>NaN</code> the resulting <code>double</code> will
    * be set to <code>NaN</code>.
    *
    * @param    value
    *           The number to round.
    *
    * @return   The rounded <code>double</code> result.
    */
   public static double round(double value)
   {
      return truncateWorker(value, decimal.MAX_SCALE, true);
   }

   /**
    * Truncates the passed <code>decimal</code> to the user-specified number
    * of digits of precision to the right side of the decimal point.  The
    * specified precision must be non-negative and any value greater than 10
    * is forced to 10.
    * <p>
    * If this instance represents the <code>unknown value</code>, the
    * result will be the <code>unknown value</code>.
    *
    * @param    value
    *           The number to truncate.
    * @param    precision
    *           The number of digits of precision to support.
    *
    * @return   The truncated <code>decimal</code> result.
    */
   public static decimal truncate(decimal value, integer precision)
   {
      return new decimal(truncate(value.toBigDecimal(),
                                  precision.intValue()));
   }
   /**
    * Truncates the passed <code>decimal</code> to the user-specified number
    * of digits of precision to the right side of the decimal point.  The
    * specified precision must be non-negative and any value greater than 10
    * is forced to 10.
    * <p>
    * If this instance represents the <code>unknown value</code>, the
    * result will be the <code>unknown value</code>.
    *
    * @param    value
    *           The number to truncate.
    * @param    precision
    *           The number of digits of precision to support.
    *
    * @return   The truncated <code>decimal</code> result.
    */
   public static decimal truncate(decimal value, int precision)
   {
      return new decimal(truncate(value.toBigDecimal(), precision));
   }
   /**
    * Truncates the passed <code>BigDecimal</code> to the specified number
    * of digits of precision to the right side of the decimal point.  The
    * specified precision must be non-negative and any value greater than 10
    * is forced to 10.
    * <p>
    * If the value is <code>null</code>, then <code>null</code> is returned.
    *
    * @param    value
    *           The number to truncate.
    * @param    precision
    *           The number of digits of precision to support.
    *
    * @return   The truncated <code>BigDecimal</code> result.
    */
   public static BigDecimal truncate(BigDecimal value, int precision)
   {
      return round(value, precision, BigDecimal.ROUND_DOWN);
   }
   /**
    * Truncates the passed <code>double</code> to the user-specified number
    * of digits of precision to the right side of the decimal point.  The
    * specified precision must be non-negative and any value greater than 10
    * is forced to 10.
    * <p>
    * If the value is <code>NaN</code> the resulting <code>double</code> will
    * be set to <code>NaN</code>.
    *
    * @param    value
    *           The number to truncate.
    * @param    precision
    *           The number of digits of precision to support.
    *
    * @return   The truncated <code>double</code> result.
    */
   public static double truncate(double value, int precision)
   {
      return truncateWorker(value, precision, false);
   }

   /**
    * Implements the unary minus operator for the <code>decimal</code> type.
    * <p>
    * If the operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    d
    *           The <code>decimal</code> to negate.
    *
    * @return   The negated result.
    */
   public static decimal negate(decimal d)
   {
      if (d.isUnknown())
         return (decimal) d.instantiateUnknown();
      
      return (new decimal(d.toBigDecimal().negate()));
   }
   /**
    * Implements the unary minus operator.  This is a "convenience" method
    * which makes it possible to use a common approach for unary minus. This
    * is needed to allow the proper wrapping to be handled in the runtime
    * rather than doing this in client code.
    *
    * @param    d
    *           The number to negate.
    *
    * @return   The negated result.
    */
   public static decimal negate(double d)
   {
      return new decimal(-d);
   }
   /**
    * Implements the unary minus operator for the <code>integer</code> type.
    * <p>
    * If the operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    i
    *           The <code>integer</code> to negate.
    *
    * @return   The negated result.
    */
   public static integer negate(integer i)
   {
      if (i.isUnknown())
         return (integer) i.instantiateUnknown();
      
      return new integer(-(i.intValue()));
   }
   /**
    * Implements the unary minus operator.  This is a "convenience" method
    * which makes it possible to use a common approach for unary minus. This
    * is needed to allow the proper wrapping to be handled in the runtime
    * rather than doing this in client code.
    *
    * @param    i
    *           The number to negate.
    *
    * @return   The negated result.
    */
   public static integer negate(int i)
   {
      return new integer(-i);
   }   /**
    * Implements the unary minus operator for the <code>integer</code> type.
    * <p>
    * If the operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    i
    *           The <code>integer</code> to negate.
    *
    * @return   The negated result.
    */
   public static int64 negate(int64 i)
   {
      if (i.isUnknown())
         return (int64) i.instantiateUnknown();
      
      return new int64(-(i.longValue()));
   }
   /**
    * Implements the unary minus operator.  This is a "convenience" method
    * which makes it possible to use a common approach for unary minus. This
    * is needed to allow the proper wrapping to be handled in the runtime
    * rather than doing this in client code.
    *
    * @param    i
    *           The number to negate.
    *
    * @return   The negated result.
    */
   public static int64 negate(long i)
   {
      return new int64(-i);
   }

   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    * <p>
    * If either operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static decimal plus(NumberType op1, NumberType op2)
   {           
      if (op1.isUnknown() || op2.isUnknown())
         return new decimal();
      
      BigDecimal d1 = (op1 instanceof decimal
                       ? ((decimal) op1).toBigDecimal()
                       : null);
      BigDecimal d2 = (op2 instanceof decimal
                       ? ((decimal) op2).toBigDecimal()
                       : null);
      
      if (d1 != null || d2 != null)
      {
         if (d1 == null)
         {
            d1 = new BigDecimal(op1.longValue());
         }
         else if (d2 == null)
         {
            d2 = new BigDecimal(op2.longValue());
         }
         
         return (new decimal(d1.add(d2)));
      }
      
      return (new decimal(op1.longValue() + op2.longValue()));
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static decimal plus(double op1, NumberType op2)
   {
      return plus(new decimal(op1), op2);
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static decimal plus(NumberType op1, double op2)
   {
      return plus(op1, new decimal(op2));
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static decimal plus(double op1, double op2)
   {
      return plus(new decimal(op1), new decimal(op2));
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    * <p>
    * If either operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static integer plus(integer op1, integer op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
         return (integer) op1.instantiateUnknown();
      
      return new integer((op1.intValue() + op2.intValue()));
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static integer plus(integer op1, int op2)
   {
      if (op1.isUnknown())
         return (integer) op1.instantiateUnknown();
      
      return new integer(op1.intValue() + op2);
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static integer plus(int op1, integer op2)
   {
      if (op2.isUnknown())
         return (integer) op2.instantiateUnknown();
      
      return new integer((op1 + op2.intValue()));
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static integer plus(int op1, int op2)
   {
      return new integer(op1 + op2);
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    * <p>
    * If either operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static int64 plus(int64 op1, int64 op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
         return (int64) op1.instantiateUnknown();
      
      return new int64((op1.longValue() + op2.longValue()));
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static int64 plus(int64 op1, long op2)
   {
      if (op1.isUnknown())
         return (int64) op1.instantiateUnknown();
      
      return new int64(op1.longValue() + op2);
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static int64 plus(long op1, int64 op2)
   {
      if (op2.isUnknown())
         return (int64) op2.instantiateUnknown();
      
      return new int64(op1 + op2.longValue());
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static int64 plus(long op1, long op2)
   {
      return new int64(op1 + op2);
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static int64 plus(int64 op1, int op2)
   {
      if (op1.isUnknown())
         return (int64) op1.instantiateUnknown();
      
      return new int64(op1.longValue() + op2);
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static int64 plus(int op1, int64 op2)
   {
      if (op2.isUnknown())
         return (int64) op2.instantiateUnknown();
      
      return new int64(op1 + op2.longValue());
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static int64 plus(integer op1, long op2)
   {
      if (op1.isUnknown())
         return (int64) op1.instantiateUnknown();
      
      return new int64(op1.longValue() + op2);
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static int64 plus(long op1, integer op2)
   {
      if (op2.isUnknown())
         return (int64) op2.instantiateUnknown();
      
      return new int64(op1 + op2.longValue());
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static int64 plus(integer op1, int64 op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
      {
         return new int64();
      }

      return new int64(op1.longValue() + op2.longValue());
   }
   /**
    * Implements the binary plus arithmetic operator to add the right
    * operand to the left and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The added result.
    */
   public static int64 plus(int64 op1, integer op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
      {
         return new int64();
      }

      return new int64(op1.longValue() + op2.longValue());
   }

   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    * <p>
    * If either operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static decimal minus(NumberType op1, NumberType op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
         return new decimal();
      
      BigDecimal d1 = (op1 instanceof decimal
                       ? ((decimal) op1).toBigDecimal()
                       : null);
      BigDecimal d2 = (op2 instanceof decimal
                       ? ((decimal) op2).toBigDecimal()
                       : null);
      
      if (d1 != null || d2 != null)
      {
         if (d1 == null)
         {
            d1 = new BigDecimal(op1.intValue());
         }
         else if (d2 == null)
         {
            d2 = new BigDecimal(op2.intValue());
         }
         
         return (new decimal(d1.subtract(d2)));
      }
      
      return (new decimal(op1.longValue() - op2.longValue()));
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static decimal minus(double op1, NumberType op2)
   {
      return minus(new decimal(op1), op2);
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static decimal minus(NumberType op1, double op2)
   {
      return minus(op1, new decimal(op2));
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static decimal minus(double op1, double op2)
   {
      return minus(new decimal(op1), new decimal(op2));
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    * <p>
    * If either operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static integer minus(integer op1, integer op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
         return (integer) op1.instantiateUnknown();
      
      return new integer((op1.intValue() - op2.intValue()));
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static integer minus(int op1, integer op2)
   {
      if (op2.isUnknown())
         return (integer) op2.instantiateUnknown();
      
      return new integer(op1 - op2.intValue());
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static integer minus(integer op1, int op2)
   {
      if (op1.isUnknown())
         return (integer) op1.instantiateUnknown();
      
      return new integer(op1.intValue() - op2);
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static integer minus(int op1, int op2)
   {
      return new integer(op1 - op2);
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    * <p>
    * If either operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static int64 minus(int64 op1, int64 op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
         return (int64) op1.instantiateUnknown();
      
      return new int64((op1.longValue() - op2.longValue()));
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static int64 minus(long op1, int64 op2)
   {
      if (op2.isUnknown())
         return (int64) op2.instantiateUnknown();
      
      return new int64(op1 - op2.longValue());
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static int64 minus(int64 op1, long op2)
   {
      if (op1.isUnknown())
         return (int64) op1.instantiateUnknown();
      
      return new int64(op1.longValue() - op2);
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static int64 minus(long op1, long op2)
   {
      return new int64(op1 - op2);
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static int64 minus(int op1, int64 op2)
   {
      if (op2.isUnknown())
         return (int64) op2.instantiateUnknown();
      
      return new int64(op1 - op2.longValue());
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static int64 minus(int64 op1, int op2)
   {
      if (op1.isUnknown())
         return (int64) op1.instantiateUnknown();
      
      return new int64(op1.longValue() - op2);
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static int64 minus(long op1, integer op2)
   {
      if (op2.isUnknown())
      {
         return new int64();
      }
      
      return new int64(op1 - op2.longValue());
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static int64 minus(integer op1, long op2)
   {
      if (op1.isUnknown())
      {
         return new int64();
      }
      
      return new int64(op1.longValue() - op2);
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static int64 minus(int64 op1, integer op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
      {
         return new int64();
      }

      return new int64(op1.longValue() - op2.longValue());
   }
   /**
    * Implements the binary minus arithmetic operator to subtract the
    * right operand from the left and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The subtracted result.
    */
   public static int64 minus(integer op1, int64 op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
      {
         return new int64();
      }

      return new int64(op1.longValue() - op2.longValue());
   }

   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    * <p>
    * If either operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static decimal multiply(NumberType op1, NumberType op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
         return (decimal) op1.instantiateUnknown();
      
      BigDecimal d1 = (op1 instanceof decimal
                       ? ((decimal) op1).toBigDecimal()
                       : null);
      BigDecimal d2 = (op2 instanceof decimal
                       ? ((decimal) op2).toBigDecimal()
                       : null);
      
      if (d1 != null || d2 != null)
      {
         if (d1 == null)
         {
            d1 = new BigDecimal(op1.longValue());
         }
         else if (d2 == null)
         {
            d2 = new BigDecimal(op2.longValue());
         }
         
         return (new decimal(d1.multiply(d2)));
      }
      
      return (new decimal(op1.longValue() * op2.longValue()));
   }
   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static decimal multiply(double op1, NumberType op2)
   {
      return multiply(new decimal(op1), op2);
   }
   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static decimal multiply(NumberType op1, double op2)
   {
      return multiply(op1, new decimal(op2));
   }
   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static decimal multiply(double op1, double op2)
   {
      return multiply(new decimal(op1), new decimal(op2));
   }
   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    * <p>
    * If either operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static integer multiply(integer op1, integer op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
         return (integer) op1.instantiateUnknown();
      
      return new integer((op1.intValue() * op2.intValue()));
   }
   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static integer multiply(int op1, integer op2)
   {
      if (op2.isUnknown())
         return (integer) op2.instantiateUnknown();
      
      return new integer(op1 * op2.intValue());
   }
   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static integer multiply(integer op1, int op2)
   {
      if (op1.isUnknown())
         return (integer) op1.instantiateUnknown();
      
      return new integer(op1.intValue() * op2);
   }
   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static integer multiply(int op1, int op2)
   {
      return new integer(op1 * op2);
   }
   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    * <p>
    * If either operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static int64 multiply(int64 op1, int64 op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
      {
         return new int64();
      }

      return new int64(op1.longValue() * op2.longValue());
   }
   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static int64 multiply(int64 op1, integer op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
      {
         return new int64();
      }

      return new int64(op1.longValue() * op2.longValue());
   }
   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static int64 multiply(integer op1, int64 op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
      {
         return new int64();
      }

      return new int64(op1.longValue() * op2.longValue());
   }
   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    * <p>
    * If either operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static int64 multiply(long op1, long op2)
   {
      return new int64(op1 * op2);
   }
   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static int64 multiply(long op1, integer op2)
   {
      if (op2.isUnknown())
      {
         return new int64();
      }

      return new int64(op1 * op2.longValue());
   }
   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static int64 multiply(integer op1, long op2)
   {
      if (op1.isUnknown())
      {
         return new int64();
      }

      return new int64(op1.longValue() * op2);
   }
   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    * <p>
    * If either operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static int64 multiply(int64 op1, long op2)
   {
      if (op1.isUnknown())
      {
         return new int64();
      }

      return new int64(op1.longValue() * op2);
   }
   /**
    * Implements the binary multiply arithmetic operator to multiply the
    * two operands and return the result.
    * <p>
    * If either operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The multiplied result.
    */
   public static int64 multiply(long op1, int64 op2)
   {
      if (op2.isUnknown())
      {
         return new int64();
      }

      return new int64(op1 * op2.longValue());
   }

   /**
    * Implements the binary floating-point division arithmetic operator to 
    * divide the left operand by the right operand and return the result.
    * There is no operator that provides integer division (use 
    * {@link #truncate} on the result if this is necessary).
    * <p>
    * If either operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.  In addition, division by 0 will
    * silently result in the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The divided result.
    */
   public static decimal divide(NumberType op1, NumberType op2)
   {
      if (op1.isUnknown() || op2.isUnknown() || op2.doubleValue() == 0.0D)
         return new decimal();
      
      BigDecimal d1 = (op1 instanceof decimal
                       ? ((decimal) op1).toBigDecimal()
                       : new decimal(op1.longValue()).toBigDecimal());
      BigDecimal d2 = (op2 instanceof decimal
                       ? ((decimal) op2).toBigDecimal()
                       : new decimal(op2.longValue()).toBigDecimal());
      
      return (new decimal(d1.divide(d2,
                                    decimal.MAX_SCALE,
                                    BigDecimal.ROUND_HALF_UP)));
   }
   /**
    * Implements the binary floating-point division arithmetic operator to 
    * divide the left operand by the right operand and return the result.
    * There is no operator that provides integer division (use 
    * {@link #truncate} on the result if this is necessary).
    * <p>
    * If the second operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.  In addition, division by 0 will
    * silently result in the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The divided result.
    */
   public static decimal divide(double op1, NumberType op2)
   {
      return divide(new decimal(op1), op2);
   }
   /**
    * Implements the binary floating-point division arithmetic operator to 
    * divide the left operand by the right operand and return the result.
    * There is no operator that provides integer division (use 
    * {@link #truncate} on the result if this is necessary).
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.  In addition, division by 0 will
    * silently result in the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The divided result.
    */
   public static decimal divide(NumberType op1, double op2)
   {
      return divide(op1, new decimal(op2));
   }
   /**
    * Implements the binary floating-point division arithmetic operator to 
    * divide the left operand by the right operand and return the result.
    * There is no operator that provides integer division (use 
    * {@link #truncate} on the result if this is necessary).
    * <p>
    * Division by 0 will silently result in the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand.
    *
    * @return   The divided result.
    */
   public static decimal divide(double op1, double op2)
   {
      return divide(new decimal(op1), new decimal(op2));
   }

   /**
    * Implements the binary modulo (remainder) arithmetic operator which
    * divides the left operand by the right operand (the base) and returns 
    * the resulting integral remainder.  The base must be > 0.
    * <p>
    * If the left operand is greater than zero, then the result will be
    * as expected from normal integral arithmetic (the remainder after
    * integer division will be returned). However, if the left operand is
    * not greater than zero, then the result will depend on the value of
    * the remainder after the absolute value of the left operand is integrally
    * divided into the base.  Zero will be returned if the remainder in this
    * case is zero, otherwise the base minus the remainder will be returned.
    * Warning: the Progress documentation only suggests that this is an
    * invalid case which should not be relied upon. However, client
    * application code has been found to be reliant upon this behavior.
    * <p>
    * If either operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand (number to divide).
    * @param    op2
    *           The right operand (the divisor or modulo base).
    *
    * @return   The remainder result.
    */
   public static int64 modulo(NumberType op1, NumberType op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
         return new int64();
      
      if (op2.longValue() < 1)
      {
         return genModuloError(op2.doubleValue());
      }
      
      // common case;  much faster than BigDecimal
      if ((op1 instanceof integer || op1 instanceof int64) && 
          (op2 instanceof integer || op2 instanceof int64))
      {
         return modulo(op1.longValue(), op2.longValue());
      }

      BigDecimal d1 = (op1 instanceof decimal
                       ? ((decimal) op1).toBigDecimal()
                       : new decimal(op1.longValue()).toBigDecimal());
      BigDecimal d2 = (op2 instanceof decimal
                       ? ((decimal) op2).toBigDecimal()
                       : new decimal(op2.longValue()).toBigDecimal());
      
      BigDecimal modVal = d1.abs().remainder(d2);
      
      // common case
      if (d1.compareTo(decimal.ZERO) >= 1)
    	   return (new int64(modVal));
      
      // handle non-positive input
      if (modVal.compareTo(decimal.ZERO) == 0)
         return new int64(0);
      
      return new int64(d2.subtract(modVal));
   }
   /**
    * Implements the binary modulo (remainder) arithmetic operator which
    * divides the left operand by the right operand (the base) and returns 
    * the resulting integral remainder.  The base must be > 0. 
    * <p>
    * If the second operand is the <code>unknown value</code>, the result
    * will be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand (number to divide).
    * @param    op2
    *           The right operand (the divisor or modulo base).
    *
    * @return   The remainder result.
    */
   public static int64 modulo(double op1, NumberType op2)
   {
      return modulo(new decimal(op1), op2);
   }
   /**
    * Implements the binary modulo (remainder) arithmetic operator which
    * divides the left operand by the right operand (the base) and returns 
    * the resulting integral remainder.  The base must be > 0. 
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand (number to divide).
    * @param    op2
    *           The right operand (the divisor or modulo base).
    *
    * @return   The remainder result.
    */
   public static int64 modulo(NumberType op1, double op2)
   {
      return modulo(op1, new decimal(op2));
   }
   /**
    * Implements the binary modulo (remainder) arithmetic operator which
    * divides the left operand by the right operand (the base) and returns 
    * the resulting integral remainder.  The base must be > 0. 
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand (number to divide).
    * @param    op2
    *           The right operand (the divisor or modulo base).
    *
    * @return   The remainder result.
    */
   public static int64 modulo(int op1, NumberType op2)
   {
      return modulo(new integer(op1), op2);
   }
   /**
    * Implements the binary modulo (remainder) arithmetic operator which
    * divides the left operand by the right operand (the base) and returns 
    * the resulting integral remainder.  The base must be > 0. 
    * <p>
    * If the first operand is the <code>unknown value</code>, the result will
    * be the <code>unknown value</code>.
    *
    * @param    op1
    *           The left operand (number to divide).
    * @param    op2
    *           The right operand (the divisor or modulo base).
    *
    * @return   The remainder result.
    */
   public static int64 modulo(NumberType op1, int op2)
   {
      return modulo(op1, new integer(op2));
   }
   /**
    * Implements the binary modulo (remainder) arithmetic operator which
    * divides the left operand by the right operand (the base) and returns 
    * the resulting integral remainder.  The base must be > 0. 
    * <p>
    * If the left operand is greater than zero, then the result will be
    * as expected from normal integral arithmetic (the remainder after
    * integer division will be returned). However, if the left operand is
    * not greater than zero, then the result will depend on the value of
    * the remainder after the absolute value of the left operand is integrally
    * divided into the base.  Zero will be returned if the remainder in this
    * case is zero, otherwise the base minus the remainder will be returned.
    * Warning: the Progress documentation only suggests that this is an
    * invalid case which should not be relied upon. However, client
    * application code has been found to be reliant upon this behavior.
    *
    * @param    op1
    *           The left operand (number to divide).
    * @param    op2
    *           The right operand (the divisor or modulo base).
    *
    * @return   The remainder result.
    */
   public static int64 modulo(double op1, double op2)
   {
      return modulo((long) op1, (long) op2);
   }
   /**
    * Implements the binary modulo (remainder) arithmetic operator which
    * divides the left operand by the right operand (the base) and returns 
    * the resulting integral remainder.  The base must be > 0. 
    * <p>
    * If the left operand is greater than zero, then the result will be
    * as expected from normal integral arithmetic (the remainder after
    * integer division will be returned). However, if the left operand is
    * not greater than zero, then the result will depend on the value of
    * the remainder after the absolute value of the left operand is integrally
    * divided into the base.  Zero will be returned if the remainder in this
    * case is zero, otherwise the base minus the remainder will be returned.
    * Warning: the Progress documentation only suggests that this is an
    * invalid case which should not be relied upon. However, client
    * application code has been found to be reliant upon this behavior.
    *
    * @param    op1
    *           The left operand (number to divide).
    * @param    op2
    *           The right operand (the divisor or modulo base).
    *
    * @return   The remainder result.
    */
   public static int64 modulo(long op1, long op2)
   {
      if (op2 < 1)
      {
         genModuloError(op2);
         return new int64();
      }
      
      long result = Math.abs(op1) % op2;
      
      // common case
      if (op1 > 0)
         return new int64(result);
      
      return result == 0 ? new int64(0) : new int64(op2 - result);
   }
   /**
    * Implements the binary modulo (remainder) arithmetic operator which
    * divides the left operand by the right operand (the base) and returns 
    * the resulting integral remainder.  The base must be > 0. 
    * <p>
    * If the left operand is greater than zero, then the result will be
    * as expected from normal integral arithmetic (the remainder after
    * integer division will be returned). However, if the left operand is
    * not greater than zero, then the result will depend on the value of
    * the remainder after the absolute value of the left operand is integrally
    * divided into the base.  Zero will be returned if the remainder in this
    * case is zero, otherwise the base minus the remainder will be returned.
    * Warning: the Progress documentation only suggests that this is an
    * invalid case which should not be relied upon. However, client
    * application code has been found to be reliant upon this behavior.
    *
    * @param    op1
    *           The left operand (number to divide).
    * @param    op2
    *           The right operand (the divisor or modulo base).
    *
    * @return   The remainder result.
    */
   public static integer modulo(int op1, int op2)
   {
      if (op2 < 1)
      {
         genModuloError(op2);
         return new integer();
      }
      
      int result = Math.abs(op1) % op2;
      
      // common case
      if (op1 > 0)
         return new integer(result);
      
      return result == 0 ? new integer(0) : new integer(op2 - result);
   }
   /**
    * Implements the binary modulo (remainder) arithmetic operator which
    * divides the left operand by the right operand (the base) and returns 
    * the resulting integral remainder.  The base must be > 0. 
    * <p>
    * If the left operand is greater than zero, then the result will be
    * as expected from normal integral arithmetic (the remainder after
    * integer division will be returned). However, if the left operand is
    * not greater than zero, then the result will depend on the value of
    * the remainder after the absolute value of the left operand is integrally
    * divided into the base.  Zero will be returned if the remainder in this
    * case is zero, otherwise the base minus the remainder will be returned.
    * Warning: the Progress documentation only suggests that this is an
    * invalid case which should not be relied upon. However, client
    * application code has been found to be reliant upon this behavior.
    *
    * @param    op1
    *           The left operand (number to divide).
    * @param    op2
    *           The right operand (the divisor or modulo base).
    *
    * @return   The remainder result.
    */
   public static integer modulo(integer op1, int op2)
   {
      if (op1.isUnknown())
      {
         return new integer();
      }
      return modulo(op1.intValue(), op2);
   }
   /**
    * Implements the binary modulo (remainder) arithmetic operator which
    * divides the left operand by the right operand (the base) and returns 
    * the resulting integral remainder.  The base must be > 0. 
    * <p>
    * If the left operand is greater than zero, then the result will be
    * as expected from normal integral arithmetic (the remainder after
    * integer division will be returned). However, if the left operand is
    * not greater than zero, then the result will depend on the value of
    * the remainder after the absolute value of the left operand is integrally
    * divided into the base.  Zero will be returned if the remainder in this
    * case is zero, otherwise the base minus the remainder will be returned.
    * Warning: the Progress documentation only suggests that this is an
    * invalid case which should not be relied upon. However, client
    * application code has been found to be reliant upon this behavior.
    *
    * @param    op1
    *           The left operand (number to divide).
    * @param    op2
    *           The right operand (the divisor or modulo base).
    *
    * @return   The remainder result.
    */
   public static integer modulo(int op1, integer op2)
   {
      if (op2.isUnknown())
      {
         return new integer();
      }
      return modulo(op1, op2.intValue());
   }
   /**
    * Implements the binary modulo (remainder) arithmetic operator which
    * divides the left operand by the right operand (the base) and returns 
    * the resulting integral remainder.  The base must be > 0. 
    * <p>
    * If the left operand is greater than zero, then the result will be
    * as expected from normal integral arithmetic (the remainder after
    * integer division will be returned). However, if the left operand is
    * not greater than zero, then the result will depend on the value of
    * the remainder after the absolute value of the left operand is integrally
    * divided into the base.  Zero will be returned if the remainder in this
    * case is zero, otherwise the base minus the remainder will be returned.
    * Warning: the Progress documentation only suggests that this is an
    * invalid case which should not be relied upon. However, client
    * application code has been found to be reliant upon this behavior.
    *
    * @param    op1
    *           The left operand (number to divide).
    * @param    op2
    *           The right operand (the divisor or modulo base).
    *
    * @return   The remainder result.
    */
   public static integer modulo(integer op1, integer op2)
   {
      if (op2.isUnknown())
      {
         return new integer();
      }
      return modulo(op1.intValue(), op2.intValue());
   }

   /**
    * Rounds the given decimal value to the user-specified number of digits
    * of precision to the right side of the decimal point, using the given
    * rounding mode.  The rounding mode determines whether the value is
    * rounded or truncated to the specified precision.
    * <p>
    * For rounding mode <code>BigDecimal.ROUND_HALF_UP</code>:  If the
    * precision is specified as X and the X+1th digit is 5 or more, the Xth
    * digit is increased by 1, otherwise it is left alone.  All digits after
    * the Xth are then dropped/truncated.
    * <p>
    * For rounding mode <code>BigDecimal.ROUND_DOWN</code>:  If the precision
    * is specified as X then all digits after the Xth are dropped/truncated.
    * <p>
    * The specified precision must be non-negative and any value greater than
    * 10 is forced to 10.
    * <p>
    * If the value is <code>null</code>, <code>null</code> is returned.  If
    * precision is negative, 0 is returned.
    *
    * @param    value
    *           The number to round.
    * @param    precision
    *           The number of digits of precision to support.
    * @param    roundingMode
    *           Either <code>BigDecimal.ROUND_HALF_UP<code> (to round) or
    *           <code>BigDecimal.ROUND_DOWN</code> (to truncate).
    *
    * @return   The rounded/truncated <code>double</code> result.
    */
   private static BigDecimal round(BigDecimal value,
                                   int precision,
                                   int roundingMode)
   {
      // negative precision is illegal
      if (precision < 0)
      {
         String err = "Decimal precision cannot be negative";
         ErrorManager.recordOrThrowError(84, err);
         
         return decimal.ZERO;
      }
      
      // return null if input is null
      if (value == null)
      {
         return null;
      }
      
      // force to decimal.MAX_SCALE digits if anything greater is set
      if (precision > decimal.MAX_SCALE)
      {
         precision = decimal.MAX_SCALE;
      }
      
      return value.setScale(precision, roundingMode);
   }
   
   /**
    * Core logic to round or truncate a passed <code>double</code> to the
    * user-specified precision (number of digits to the right of the
    * decimal point). If <code>round</code> is <code>true</code> and if the 
    * precision is specified as X then when the X+1th digit is 5 or more, the
    * Xth digit is increased by 1, otherwise it is left alone. All digits
    * after the Xth are then dropped/truncated.  If <code>round</code> is
    * <code>false</code> then the Xth digit is always unchanged and all
    * digits greater than X are truncated. The specified precision
    * must be non-negative and any value greater than 10 is forced to 10.
    * <p>
    * The algorithm used relies upon scaling (multiplying) the input value
    * by a power of 10 based on the precision (for example, by
    * <code>Math.pow(10, precision)</code>).  This leaves all digits to be
    * kept in the result, to the left of the decimal point.  Then this
    * <code>double</code> is cast to a <code>long</code> (truncate mode)
    * or rounded into a <code>long</code> (round mode) which drops the
    * extra digits.  Finally, the value is de-scaled (divided by) the same
    * scaling value to obtain a properly scaled result. 
    * <p>
    * This algorithm will fail when used on large numbers (positive or
    * negative) with large precision.  This is due to the fact that the 
    * intermediate resulting <code>double</code> (after scaling) can be larger
    * than <code>Long.MAX_VALUE</code> or smaller than
    * <code>Long.MIN_VALUE</code> which breaks the algorithm.
    * <p>
    * If the passed value to be rounded/truncated is <code>NaN</code>, the
    * result will be <code>NaN</code>.
    *
    * @param    value
    *           The value to be rounded or truncated.
    * @param    precision
    *           The number of significant digits to maintain to the right of
    *           the decimal point.  Use 0 to get a value that is integral.
    * @param    round
    *           <code>true</code> to using rounding, <code>false</code> to
    *           obtain truncation.
    *
    * @return   The rounded or truncated result.
    *
    * @throws   IllegalArgumentException
    *           If the intermediate result is too large or too small to fit
    *           into a <code>long</code> (e.g. positive or negative 
    *           infinity) or if the precision is negative.
    */
   private static double truncateWorker(double  value,
                                        int     precision,
                                        boolean round)
   throws IllegalArgumentException
   {
      // negative precision is illegal
      if (precision < 0)
      {
         String err = "Decimal precision cannot be negative";
         ErrorManager.recordOrThrowError(84, err); 
         return 0;
      }
      
      // return NaN if input is NaN
      if (Double.isNaN(value))
      {
         // nothing to do!
         return value;
      }
      
      // force to 10 digits if anything greater is set
      if (precision > decimal.MAX_SCALE)
      {
         precision = decimal.MAX_SCALE;
      }
      
      double scale = 10000000000.0D;
      
      if (precision != decimal.MAX_SCALE)
      {
         scale = Math.pow(decimal.MAX_SCALE, precision);
      }
      
      // the left side of the decimal point should never change BUT these
      // extra significant digits can change how numbers are represented
      // in floating point format; since these digits are not needed for
      // the rounding or truncation operation, we safe them off here and
      // will add them back later; this allows the calculation to be done
      // on a double in which we only ever care about 11 significant digits 
      // (those digits to the left of the decimal point which determine the
      // result)
      long left = (long) value;
      double right = value - left;
      
      // "shift" the right side value such that any digits that we don't want
      // are on the right side of the decimal point and all digits we must
      // save are on the left side of decimal point
      double scaled = right * scale;
      
      // a Java double can only have a total of ~16 significant digits 
      // (on the left side of the decimal point), so if the left side has
      // 16 or more digits, the result MAY SILENTLY lose or drop
      // least significant digits and the resulting value will be stored in
      // the closest possible representation; since Java doesn't notify us
      // when this occurs, we compare it to the maximum left side size
      // this allows us to detect the case where the result will be wrong
      boolean loss = left > MAX_SAFE;
      
      // make sure we don't overflow/underflow or lose significant digits
      if (scaled > Long.MAX_VALUE || scaled < Long.MIN_VALUE || loss)
      {
         // alternate (slower) approach which is guaranteed to work
         
         // rounding mode (ROUND_DOWN is truncation) 
         int rmode = round ? BigDecimal.ROUND_HALF_UP : BigDecimal.ROUND_DOWN;
         
         // BD is immutable, create the full valued BD
         BigDecimal val = new BigDecimal(value);
         
         // obtain a new BD with the correct precision based on the requested
         // rounding/truncation mode
         BigDecimal res = val.setScale(precision, rmode);
         
         return res.doubleValue();
      }
      
      long trunc = 0;
      
      // remove the unneeded digits by converting to long
      if (round)
      {
         // use the rounding function to find the correct long value
         trunc = Math.round(scaled); 
      }
      else
      {
         // regular truncation
         trunc = (long) scaled;
      }
      
      // convert back to the proper scale and return the resulting
      // truncated/rounded double + the left side value      
      return trunc/scale + left;
   }
   
   /**
    * Generate an error if the 2nd operand to {@link #modulo} is not positive.
    *
    * @param    op2
    *           The second parameter (provided for insertion into the error
    *           message.
    *
    * @return   The <code>unknown value</code>.
    */
   private static int64 genModuloError(double op2)
   {
      String spec = "The second argument to MOD must be positive." +
                    " %f is invalid";
      String err  = String.format(spec, op2);
      
      ErrorManager.recordOrThrowError(89, err);
       
      return new int64();
   }
}
